import sys
import argparse

from acdc_nn import util
from acdc_nn import nn

import numpy as np
import pandas as pd

from pkg_resources import resource_filename
import functools


def load_nn(weight_path):
	num_H=[32,16]
	d=0.2
	model1, model2 = nn.build_acdc_3d(num_H,d,25)
	weight_path = resource_filename('acdc_nn', 'weights/full_dataset_TL')
	model1.load_weights(weight_path)
	return model1, model2

@functools.lru_cache(100) # memory cache to avoid reloading the same protein data multiple times
def	load_prot(profile_path, pdb_path, chain):
	r = dict(zip(['structure', 'pchain', 'seq', 'd_seq2pdb', 'd_pdb2seq'], util.pdb2info(pdb_path, chain)))
	r['prof'] = util.getProfile(profile_path)
	r['profile_path'] = profile_path
	r['pdb_path'] = pdb_path
	return r

def compute_nn_input(mut, prot):
	kvar = (mut[0], prot['d_pdb2seq'][mut[1:-1]], mut[-1]) # position in the sequence
	kvar_pdb = (mut[0], mut[1:-1], mut[-1]) # position in the pdb
	from_residue = prot['seq'][int(kvar[1]) - 1]
	if kvar[0] != from_residue:
		print('WARNING: mutation {} does not correspond to residue {} from {}'.format(kvar[0], from_residue, prot['pdb_path']))

	dist_neigh_3d = util.get_neigh_ps(kvar_pdb, 5, prot['d_seq2pdb'], prot['pchain']) 
	list_dist_neigh_3d = dist_neigh_3d[kvar]

	# extracting features
	codif = util.getMutCod(mut)
	all_profile = util.Unified_prof(kvar[1], prot['prof'], prot['seq'], list_dist_neigh_3d)
	#all_data=[*codif,*all_profile,*np.zeros(600-len(all_profile))]
	return pd.DataFrame([*codif, *all_profile, *np.zeros(600 - len(all_profile))]).T

def predict(model, wt_mutation, wt_profile, wt_structure, wt_chain, mt_mutation, mt_profile, mt_structure, mt_chain):
		wt_prot = load_prot(wt_profile, wt_structure, wt_chain)
		to_predict_dir = compute_nn_input(wt_mutation, wt_prot)
		if mt_mutation is None: # compute input features for the inverse mutation inverting the input features for the direct mutation
			to_predict_inv = to_predict_dir.copy()
			to_predict_inv.iloc[:, :20] = to_predict_inv.iloc[:, :20].replace([1.0, -1.0],[-1.0, 1.0])
		else: # compute input features for the inverse mutation using the mutated protein
			mt_prot = load_prot(mt_profile, mt_structure, mt_chain)
			to_predict_inv = compute_nn_input(mt_mutation, mt_prot)
	
		# Making input in the proper shape 
		Xm_d, X1D_d, X3D_d = nn.mkInp_3d(np.asarray(to_predict_dir).astype(np.float32), 500)
		Xm_i, X1D_i, X3D_i = nn.mkInp_3d(np.asarray(to_predict_inv).astype(np.float32), 500)

		prediction = model.predict([X3D_d, X1D_d, Xm_d , X3D_i, X1D_i, Xm_i])
		return prediction[0][0][0]
	

def main():
	parser = argparse.ArgumentParser(description='Predict the DDG of point mutations.', epilog='''
	Notes:
	Mutations are written as XNY, meaning that the residue X at position N changes to Y. X and Y are given as a one letter amino acid code and 
	N is 1-based and referred to the the PDB numbering of the relevant chain, and not the position on the sequence.
	PDB and profile files will be automatically decompressed (by gzip) if the paths end with ".gz".
	The batch file is a tab-separated table with a row for each mutation. Two row formats are possible, either the four field:
	MUT PROF PDB CHAIN
	or the eight field:
	MUT WT-PROF WT-PDB WT-CHAIN INV-MUT MT-PROF MT-PDB MT-CHAIN
	''')
	parser.add_argument('--output', metavar='OUTFILE', 
		type=argparse.FileType('w'), default=sys.stdout, 
		help='output file path, use standard output otherwise')
	parser.add_argument('--weights', metavar='NN-WEIGHTS', 
		default=resource_filename('acdc_nn', 'weights/full_dataset_TL'), 
		help='path to the network weights, in tensorflow format') # FIXME what is the name of the format?
	subparsers = parser.add_subparsers(metavar='COMMAND', dest='command', title='COMMAND')

	single_parser = subparsers.add_parser('single', help='predict a single mutation')
	single_parser.add_argument('wt_mutation', metavar='MUT', help='mutation (e.g. Q339N)')
	single_parser.add_argument('wt_profile', metavar='PROF', help='profile file path')
	single_parser.add_argument('wt_structure', metavar='PDB', help='PDB file path')
	single_parser.add_argument('wt_chain', metavar='CHAIN', 
		help='ID of the PDB chain where the mutation occurs')

	inverse_parser = subparsers.add_parser('inverse', 
		help='predict a single mutation using also the mutated protein information')
	inverse_parser.add_argument('wt_mutation', metavar='MUT', help='direct mutation (e.g. Q339N)')
	inverse_parser.add_argument('wt_profile', metavar='WT-PROF', 
		help='profile file path for the wild-type protein')
	inverse_parser.add_argument('wt_structure', metavar='WT-PDB', 
		help='PDB path for the wild-type protein')
	inverse_parser.add_argument('wt_chain', metavar='WT-CHAIN', 
		help='ID of the wild-type PDB chain where the mutation occurs')
	inverse_parser.add_argument('mt_mutation', metavar='INV-MUT', help='inverse mutation (e.g. N339Q)')
	inverse_parser.add_argument('mt_profile', metavar='MT-PROF', help='profile file path for the mutated protein')
	inverse_parser.add_argument('mt_structure', metavar='MT-PDB', 
		help='PDB file path for the mutated protein')
	inverse_parser.add_argument('mt_chain', metavar='MT-CHAIN', 
		help='ID of the mutated PDB chain where the mutation occurs')

	batch_parser = subparsers.add_parser('batch', help='predict a list of mutations from a file')
	batch_parser.add_argument('mutation_table', metavar='MUT-FILE', type=argparse.FileType('r'), 
		help='tab-separated mutation table, see later for format guidelines')

	#profile_parser = subparsers.add_parser('profile', help='compute the profile from multiple alignment file') # we could also simply accept a multialn file and make a profile
	#profile_parser.add_argument('multialn', metavar='ALN-FILE', type=argparse.FileType('r'))

	args = parser.parse_args()

	if args.command is None:
		parser.error('the following arguments are required: COMMAND')
	
	base_arg_list = ['wt_mutation', 'wt_profile', 'wt_structure', 'wt_chain', 'mt_mutation', 'mt_profile', 'mt_structure', 'mt_chain']

	model1, model2 = load_nn(args.weights) # needed by all command
	if args.command in {'single', 'inverse'}:
		args_dict = vars(args)
		mutation_args = [args_dict.get(a) for a in base_arg_list]
		ddg = predict(model1, *mutation_args)
		print(ddg, file=args.output)
	elif args.command == 'batch':
		muts = pd.read_csv(args.mutation_table, sep='\t', names=base_arg_list)
		for row, mut in muts.iterrows():
			# check na values
			if mut[base_arg_list[:4]].count() < 4:
				print('ERROR at line {}: missing values in the first four columns of batch file'.format(row + 1), file=sys.stderr)
				sys.exit(1)
			if mut[base_arg_list[4:]].count() not in {0, 4}:
				print('ERROR at line {}: columns 4-8 of the batch file must be either all missing or all present'.format(row + 1), file=sys.stderr)
				sys.exit(1)
					
			mutation_args = [None if pd.isnull(v) else v for v in mut.values]
			ddg = predict(model1, *mutation_args)
			print(ddg, file=args.output)
		
	#elif args.command == 'profile':
	else:
		raise NotImplementedError('Command {} not implemented'.format(args.command))

if __name__ == '__main__':
	main()



from . import acdc_nn
from . import util
import ddgun

import click
from warnings import warn
import functools

class Substitution(click.ParamType):
	'''Click parameter class for substitutions'''
	name = 'amino acid substitution'
	def convert(self, value, param, ctx):
		if isinstance(value, ddgun.Substitution):
			return value
		try:
			return ddgun.Substitution.parse(value)
		except Exception as e:
			self.fail(f"{value!r} is not a valid {self.name}", param, ctx)

help_notes = '''Notes:
Mutations are written as XNY, meaning that the residue X at position N changes to Y. X and Y are given as a one letter amino acid code and
N is 1-based and refers to the the PDB numbering of the relevant chain, and not the position on the sequence.
PDB and profile files will be automatically decompressed (by gzip) if the paths end with ".gz".
'''

@click.group(epilog=help_notes)
def cli():
	pass

@cli.command(epilog=help_notes)  # TODO add option for the weights
@click.argument("sub", type=Substitution())
@click.argument("profile", type=click.Path(exists=True, readable=True))  # FIXME use File object
def seq(sub, profile):
	'''Predict DDG of SUB from the protein PROFILE.

\b
SUB is an amino acid substitution (e.g. Q339P).
PROFILE is the path to a protein profile file.

Uses a trained ACDC-NN Seq that does not require protein structural information.'''
	wt_prof = ddgun.Profile(profile)
	net = acdc_nn.ACDCSeq()
	ddg = net.predict(sub, wt_prof)
	click.echo(ddg)

@cli.command(epilog=help_notes)
@click.argument("sub", type=Substitution())
@click.argument("profile", type=click.Path(exists=True, readable=True))
@click.argument("pdb", type=click.Path(exists=True, readable=True))
@click.argument("chain")
#@click.option('--inverse', type=(Substitution(), click.Path(exists=True, readable=True), click.Path(exists=True, readable=True), str), help=')
def struct(sub, profile, pdb, chain):  #FIXME add inverse mut 
	'''Predict DDG of SUB from the protein PROFILE and PDB structure.

\b
SUB is an amino acid substitution (e.g. Q339P).
PROFILE is the path to a protein profile file.
PDB is the path to the protein structure in PDB file.
CHAIN is the PDB chain to be used.

Uses a trained ACDC-NN that requires protein structural information.'''
	wt_prof = util.getProfile(profile)  #FIXME use Profile
	wt_struct = acdc_nn.Structure(pdb, chain)
	net = acdc_nn.ACDC3D()
	ddg = net.predict(str(sub), wt_prof, wt_struct)
	click.echo(ddg)

@cli.command(epilog=help_notes)
@click.argument("sub", type=Substitution())
@click.argument("profile", type=click.Path(exists=True, readable=True))
@click.argument("pdb", type=click.Path(exists=True, readable=True))
@click.argument("chain")
@click.argument("isub", type=Substitution())
@click.argument("iprofile", type=click.Path(exists=True, readable=True))
@click.argument("ipdb", type=click.Path(exists=True, readable=True))
@click.argument("ichain")
#@click.option('--inverse', type=(Substitution(), click.Path(exists=True, readable=True), click.Path(exists=True, readable=True), str), help=')
def istruct(sub, profile, pdb, chain):  #FIXME add inverse mut 
	'''Predict DDG using both the wild-type and mutated protein structures.

\b
SUB is an amino acid substitution (e.g. Q339P).
PROFILE is the path to a protein profile file.
PDB is the path to the protein structure in PDB file.
CHAIN is the PDB chain to be used.
ISUB, IPROFILE, IPDB and ICHAIN are the same for the mutated protein.

Uses a trained ACDC-NN that requires protein structural information.'''
	wt_prof = util.getProfile(profile)  #FIXME use Profile
	wt_struct = acdc_nn.Structure(pdb, chain)
	net = acdc_nn.ACDC3D()
	ddg = net.predict(str(sub), wt_prof, wt_struct)
	click.echo(ddg)


# caching for functions
@functools.lru_cache(2)
def load_nn(seq):
	return acdc_nn.ACDCSeq() if seq else acdc_nn.ACDC3D()

@functools.lru_cache
def load_prot_seq(profile):
	return ddgun.Profile(profile)

@functools.lru_cache
def load_prot_3d(profile, pdb, chain):
	return util.getProfile(profile), acdc_nn.Structure(pdb, chain) #FIXME use Profile


@cli.command(epilog=help_notes)
@click.argument("subs", type=click.File())
def batch(subs):
	'''Predict DDG of SUBS using available information.

SUBS is a table containing one amino acid substitution per row and paths to protein profiles and optionally protein structure.

\b
Each row can have 2, 4 or 8 fields of tab separated values that are interpreted with the following schema. 
# \tPredictor   \tFields
2 \tACDC-NN Seq \tSUB PROFILE
4 \tACDC-NN     \tSUB PROFILE PDB CHAIN
8 \tACDC-NN     \tWT-SUB WT-PROFILE WT-PDB WT-CHAIN MT-SUB MT-PROFILE MT-PDB MT-CHAIN
For rows with 2 fields, that is, without structural information, the sequence-based ACDC-NN Seq predictor is used. For rows with 4 or 8 fields, the structure-based ACDC-NN is used. Outputs one DDG value for each table row.
'''
	for row, line in enumerate(subs, 1):
		fields = line.rstrip('\n\r').split('\t')
		
		# detect how many fields are available
		null_fields = [f in {'', '.', 'NA', 'na'} for f in fields]
		for i, nf in enumerate(null_fields):
			if nf:
				n = i
				break
		else:
			n = len(fields)

		if n == 2: # only profile
			seq = True
			pargs = ddgun.Substitution.parse(fields[0]), load_prot_seq(fields[1])
		elif n == 4 or n == 8: # also structure
			seq = False
			pargs = (
					str(ddgun.Substitution.parse(fields[0])),
					*load_prot_3d(*fields[1:4]))
			if n == 8: # also inverse substitution
				pargs = (
						*pargs,
						str(ddgun.Substitution.parse(fields[4])), 
						*load_prot_3d(*fields[5:8])) 
		else:
			raise ValueError(f"found {n} fields at line {row}: fields must be 2, 4 or 8")

		# check that there
		for i, nf in enumerate(null_fields[n:], n):
			if not nf:
				warn(f"found value at column {i} after missing value at column {n}, at line {row}")

		ddg = load_nn(seq=seq).predict(*pargs)

		click.echo(ddg)




from silence_tensorflow import silence_tensorflow
silence_tensorflow()

#from tensorflow import keras
from tensorflow.keras import layers
#from tensorflow.keras.layers import Lambda, Dense, Input, Dropout, Activation, Flatten, BatchNormalization, Dropout
from tensorflow.keras.models import Sequential, Model
#from tensorflow.keras import backend as K

import math



def diff(vects):
    x, y = vects
    return (x-y)/2.0

def diff_shape(shapes):
    shape1, shape2 = shapes
    return (shape1[0], 1)
def diff_shape(shapes):
    shape1, shape2 = shapes
    return (shape1[0], 1)

def avout(vects):
    x, y = vects
    return (x+y)/2.0
def avout(vects):
    x, y = vects
    return (x+y)/2.0
    


def avout_shape(shapes):
    shape1, shape2 = shapes
    return (shape1[0], 1)

def abs_loss(y_true, y_pred):
    return K.mean(K.abs(y_pred), axis=-1)

def mkInp_3d(X, col_3d):
    Xm = X[:, 0:20]
    X1D = X[:, 20:120].reshape(X.shape[0], 100//20, 20)
    X3D = X[:, 120:].reshape(X.shape[0], col_3d//20, 20)
    return Xm, X1D, X3D    

def mkInp_seq(X):
    Xm = X[:,0:20]
    X1D = X[:,20:160].reshape(X.shape[0],140//20,20)
    return Xm, X1D

def build_acdc_seq(num_H, d):
    '''Return'''
    Coding = 20 # size of the coding, 20 residues, 
    #
    num_Env1D = 7   # number of sequence neighbours
    num_Env1D_i = 7 # number of sequence neighbours
    num_IMut = Coding
    num_IMut_i = Coding
    # Create an input for the Sequence neighbours
    input_env1D = layers.Input(shape=(num_Env1D,Coding), name="input_env1D") # direct
    input_env1D_i = layers.Input(shape=(num_Env1D_i,Coding), name="input_env_i1D") # inverse
    #
    #crere layers.InputMut(V)
    inputMut = layers.Input(shape=(num_IMut,), name="inputMut")#dir
    inputMut_i = layers.Input(shape=(num_IMut_i,), name="inputMut_i")#inv
    #conv!D
    
    conv_1D = layers.Conv1D(filters=Coding, kernel_size=1, strides=1, use_bias=False, name="net_envConv1D")
    #
    net_env1D = conv_1D(input_env1D) # direct
    net_env1D_i = conv_1D(input_env1D_i) # inverse
    #
    # create the input for the mutation for 1D
    #uuuuu
    dot_prod1D = layers.Dot(axes=(2,1))([net_env1D,inputMut])
    dot_prod1D_i = layers.Dot(axes=(2,1))([net_env1D_i,inputMut_i])
    #
    conc_dir = layers.concatenate([dot_prod1D, inputMut]) ###
    conc_inv = layers.concatenate([dot_prod1D_i, inputMut_i])###
    
    net = Sequential(name="TopLayer")
    
    i = 1
    for hidden in num_H:
        net.add(layers.Dense(hidden, activation="relu", name="Dense_"+str(i)))
        net.add(layers.Dropout(d)) #0.35
        i+=1
    net.add(layers.Dense(1, activation='linear', name="Dense_"+str(i)))
    
    
    net_dir = net(conc_dir)
    # inverse
    
    net_inv = net(conc_inv)
    # difference and average layers
    out_d = layers.Lambda(diff, output_shape=diff_shape, name="out_diff")([net_dir, net_inv])
    out_av = layers.Lambda(avout, output_shape=avout_shape, name="out_av")([net_dir, net_inv])
    
    net_dir_inv = Model([ input_env1D, inputMut, input_env1D_i, inputMut_i],[out_d, out_av])
    net_dir_inv.compile(optimizer='Adam',
              loss=['logcosh', abs_loss],
              loss_weights=[1., 1.] )
    return net_dir_inv



# Build the network with global average.
def build_acdc_3d(num_H, d, num_3d):
    Coding = 20 # size of the coding, 20 residues, 
    #
    num_Env3D = num_3d   # number of 3D neighbours
    num_Env3D_i = num_3d   # number of 3D neighbours
    num_Env1D = 5   # number of sequeunce neighbours
    num_Env1D_i = 5 # number of sequence neighbours
    num_IMut = Coding
    num_IMut_i = Coding
    #
    # Create an input for the 3D neighbours
    input_env3D = layers.Input(shape=(num_Env3D,Coding), name="input_env3D") # direct 
    input_env3D_i = layers.Input(shape=(num_Env3D_i,Coding), name="input_env_i3D") # inverse
    #
    conv_3D = layers.Conv1D(filters=Coding, kernel_size=1, strides=1, use_bias=False, name="net_envConv3D")
    #
    net_env3D = conv_3D(input_env3D) # direct 
    net_env3D_i = conv_3D(input_env3D_i) # inverse
    # create the input for the mutation for 3D
    # direct
    inputMut = layers.Input(shape=(num_IMut,), name="inputMut")
    pool = layers.GlobalAveragePooling1D()(net_env3D)
    dot_prod = layers.Dot(axes=(1,1))([pool,inputMut]) 
    # inverse
    inputMut_i = layers.Input(shape=(num_IMut_i,), name="inputMut_i")
    pool_i = layers.GlobalAveragePooling1D()(net_env3D_i)
    dot_prod_i = layers.Dot(axes=(1,1))([pool_i,inputMut_i])
    # Create an input for the Sequence neighbours
    input_env1D = layers.Input(shape=(num_Env1D,Coding), name="input_env1D") # direct
    input_env1D_i = layers.Input(shape=(num_Env1D_i,Coding), name="input_env_i1D") # inverse
    #
    conv_1D = layers.Conv1D(filters=Coding, kernel_size=1, strides=1, use_bias=False, name="net_envConv1D")
    #
    net_env1D = conv_1D(input_env1D) # direct
    net_env1D_i = conv_1D(input_env1D_i) # inverse
    #
    # create the input for the mutation for 1D
    dot_prod1D = layers.Dot(axes=(2,1))([net_env1D,inputMut])
    dot_prod1D_i = layers.Dot(axes=(2,1))([net_env1D_i,inputMut_i])
    #
    # flat dense layer after the convolutional
    net = Sequential(name="TopLayer")
    #num_H=[32,16] # hidden nodes for each layer starting from the bottom
    i = 1
    for hidden in num_H:
        net.add(layers.Dense(hidden, activation="relu", name="Dense_"+str(i)))
        net.add(layers.Dropout(d)) #0.2
        i+=1
    net.add(layers.Dense(1, activation='linear', name="Dense_"+str(i)))
    # Concatenate 3D and 1D inputs
    # direct
    conc_dir = layers.concatenate([dot_prod,dot_prod1D, inputMut]) ###
    net_dir = net(conc_dir)
    # inverse
    conc_inv = layers.concatenate([dot_prod_i,dot_prod1D_i, inputMut_i])###
    net_inv = net(conc_inv)
    # difference and average layers
    out_d = layers.Lambda(diff, output_shape=diff_shape, name="out_diff")([net_dir, net_inv])
    out_av = layers.Lambda(avout, output_shape=avout_shape, name="out_av")([net_dir, net_inv])
    # final training model
    net_dir_inv = Model([input_env3D, input_env1D, inputMut, input_env3D_i, input_env1D_i, inputMut_i],[out_d, out_av])
    # partial model, to use only one flow either direct or inverse
    net_DDG = Model([input_env3D, input_env1D, inputMut],[net_dir])
    net_dir_inv.compile(optimizer='adam',
              loss=['logcosh', abs_loss],
              loss_weights=[1., 1.] )
    return net_dir_inv, net_DDG





import sys
import argparse

from . import util
from . import nn

import numpy as np
import pandas as pd

from pkg_resources import resource_filename

#@functools.lru_cache(100) # memory cache to avoid reloading the same protein data multiple times
def	load_prot(profile_path, pdb_path=None, chain=None):
	if pdb_path is None:
		r = {}
	else:
		assert chain is not None
		r = dict(zip(['structure', 'pchain', 'seq', 'd_seq2pdb', 'd_pdb2seq'], util.pdb2info(pdb_path, chain)))
	r['profile'] = util.getProfile(profile_path)
	r['pprofile'] = util.load_profile(profile_path)
	if False:
		try:
			r['pprofile'] = pd.read_csv(profile_path, sep='\t', index_col='POS')
		except ValueError:
			r['pprofile'] = pd.read_csv(profile_path, sep=' ', index_col='POS')
	r['profile_path'] = profile_path
	r['pdb_path'] = pdb_path
	return r


class ACDCSeq:
	def __init__(self, weights=True):
		self.nn = nn.build_acdc_seq(num_H=[128, 64], d=0.35)
		if weights is False:
			pass
		elif weights is True:
			self.load_weights()
		else:
			self.load_weights(weights)
		
	def load_weights(self, path=None):
		if path is None:
			path = resource_filename('acdc_nn', 'weights/acdc_nn_seq.h5')
		self.nn.load_weights(path)

	def predict(self, wt_mut, wt_prof): # TODO accept mt_prot
		codif = util.getMutCod(str(wt_mut))
		
		# local profile
		#pos = int(wt_mut[1:-1])
		wt_prof_ctx = wt_prof.get_context(wt_mut.aa_pos, context=3)
		#profile = util.profile_context(wt_prof, pos, 3)
		if False:
			profile_old = np.array(util.prof(wt_mut, wt_prot['profile'], wt_prot['profile'].keys()))
			if not np.array_equal(profile, profile_old):
				print(profile.shape)
				print(profile)
				print(profile_old.shape)
				print(profile_old)
				assert False

		# direct mutation input preparation
		To_predict_dir = pd.DataFrame([*codif,*wt_prof_ctx]).T
		Xm_d, X1D_d = nn.mkInp_seq(np.asarray(To_predict_dir).astype(np.float32))

		# inverse mutation input preparation
		To_predict_inv = To_predict_dir.copy()
		To_predict_inv.iloc[:,:20]=To_predict_inv.iloc[:,:20].replace([1.0,-1.0],[-1.0,1.0])
		Xm_i, X1D_i = nn.mkInp_seq(np.asarray(To_predict_inv).astype(np.float32))

		prediction=self.nn.predict([X1D_d, Xm_d, X1D_i, Xm_i])
		return prediction[0][0][0]

class Structure:
	def __init__(self, path, chain):
		names = ['structure', 'pchain', 'seq', 'd_seq2pdb', 'd_pdb2seq']
		values = util.pdb2info(path, chain)
		for n, v in zip(names, values):
			setattr(self, n, v)
		self.pdb_path = path
		self.pdb_chain = chain
		# FIXME maybe pdb_chain is the same as pchain
		
def prepare_3d_input(mut, prof, struct):
	kvar = (mut[0], struct.d_pdb2seq[mut[1:-1]], mut[-1]) # position in the sequence
	kvar_pdb = (mut[0], mut[1:-1], mut[-1]) # position in the pdb
	from_residue = struct.seq[int(kvar[1]) - 1]
	if kvar[0] != from_residue:
		print('WARNING: mutation {} does not correspond to residue {} from {}'.format(kvar[0], from_residue, struct.pdb_path)) # FIXME use logging

	dist_neigh_3d = util.get_neigh_ps(kvar_pdb, 5, struct.d_seq2pdb, struct.pchain)
	list_dist_neigh_3d = dist_neigh_3d[kvar]

	# extracting features
	codif = util.getMutCod(mut)
	all_profile = util.Unified_prof(kvar[1], prof, struct.seq, list_dist_neigh_3d)
	#all_data=[*codif,*all_profile,*np.zeros(600-len(all_profile))]
	return pd.DataFrame([*codif, *all_profile, *np.zeros(600 - len(all_profile))]).T

class ACDC3D:
	def __init__(self, weights=True):
		self.nn, _ = nn.build_acdc_3d(num_H=[32, 16], d=0.2, num_3d=25)
		if weights is False:
			pass
		elif weights is True:
			self.load_weights()
		else:
			self.load_weights(weights)

	def load_weights(self, path=None):
		if path is None:
			path = resource_filename('acdc_nn', 'weights/full_dataset_TL')
		self.nn.load_weights(path)

	def predict(self, wt_mut, wt_prof, wt_struct, mt_mut=None, mt_prof=None, mt_struct=None):
		direct_nn_input = prepare_3d_input(wt_mut, wt_prof, wt_struct)
		
		if mt_mut is None: # compute input features for the inverse mutation inverting the input features for the direct mutation
			inverse_nn_input = direct_nn_input.copy()
			inverse_nn_input.iloc[:, :20] = inverse_nn_input.iloc[:, :20].replace([1.0, -1.0],[-1.0, 1.0])
		else: # compute input features for the inverse mutation using the mutated protein
			inverse_nn_input = prepare_3d_input(mt_mut, mt_prof, mt_struct)
		
		# Reshaping input
		Xm_d, X1D_d, X3D_d = nn.mkInp_3d(np.asarray(direct_nn_input).astype(np.float32), 500)
		Xm_i, X1D_i, X3D_i = nn.mkInp_3d(np.asarray(inverse_nn_input).astype(np.float32), 500)

		prediction = self.nn.predict([X3D_d, X1D_d, Xm_d , X3D_i, X1D_i, Xm_i])
		return prediction[0][0][0]

	
def run_tests():
	# from acdc_nn.acdc_nn import *
	#p=load_prot('tests/profiles/2ocjA.prof.gz')
	p=load_prot('acdc-nn/tests/profiles/2ocjA.prof.gz')
	#lp = util.profile_context(p['pprofile'], 2, 3)
	#lp1 = util.prof('_3_', p['profile'], p['profile'].keys())
	print(p.keys())

	print('ACDC-NN Sequence tests')
	nn = ACDCSeq()
	nn.load_weights()
	print(nn.predict('A1C', p))

	print('ACDC-NN Structure tests')
	nn = ACDC3D()
	nn.load_weights()
	print('OK')
if __name__ == '__main__':
	run_tests()
